from snowddl.blueprint import OutboundShareBlueprint, Grant, build_grant_name_ident
from snowddl.resolver.abc_resolver import AbstractResolver, ResolveResult, ObjectType


class OutboundShareResolver(AbstractResolver):
    skip_on_empty_blueprints = True

    def get_object_type(self) -> ObjectType:
        return ObjectType.SHARE

    def get_existing_objects(self):
        existing_objects = {}

        cur = self.engine.execute_meta("SHOW SHARES")

        for r in cur:
            if r["kind"] != "OUTBOUND":
                continue

            if r["owner"] != self.engine.context.current_role:
                continue

            # Remove organization and account prefix, shares are referred in SQL by name only
            full_name = r["name"].split(".")[-1]

            existing_objects[full_name] = {
                "share": full_name,
                "database": r["database_name"],
                "accounts": r["to"].split(",") if r["to"] else [],
                "comment": r["comment"] if r["comment"] else None,
            }

        return existing_objects

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(OutboundShareBlueprint)

    def create_object(self, bp: OutboundShareBlueprint):
        query = self.engine.query_builder()

        query.append(
            "CREATE SHARE {full_name:i}",
            {
                "full_name": bp.full_name,
            },
        )

        if bp.comment:
            query.append_nl(
                "COMMENT = {comment}",
                {
                    "comment": bp.comment,
                },
            )

        self.engine.execute_unsafe_ddl(query, condition=self.engine.settings.execute_outbound_share)

        for bp_grant in bp.grants:
            self.create_grant(bp.full_name, bp_grant)

        if bp.accounts:
            self.set_accounts(bp)

        return ResolveResult.CREATE

    def compare_object(self, bp: OutboundShareBlueprint, row: dict):
        result = ResolveResult.NOCHANGE

        existing_grants = self.get_existing_share_grants(bp.full_name)

        for bp_grant in bp.grants:
            if bp_grant not in existing_grants:
                self.create_grant(bp.full_name, bp_grant)
                result = ResolveResult.GRANT

        for ex_grant in existing_grants:
            if ex_grant not in bp.grants:
                self.drop_grant(bp.full_name, ex_grant)
                result = ResolveResult.GRANT

        # SHOW SHARES command returns only 3 accounts for each OUTBOUND SHARE
        # If you have to set more accounts, it will work, but SnowDDL will be forced to run SET ACCOUNTS every time
        if bp.accounts and [str(a) for a in bp.accounts] != row["accounts"]:
            self.set_accounts(bp)
            result = ResolveResult.ALTER

        if bp.comment != row["comment"]:
            self.engine.execute_safe_ddl(
                "ALTER SHARE {full_name:i} SET COMMENT = {comment}",
                {
                    "full_name": bp.full_name,
                    "comment": bp.comment,
                },
                condition=self.engine.settings.execute_outbound_share,
            )

            result = ResolveResult.ALTER

        return result

    def drop_object(self, row: dict):
        self.engine.execute_unsafe_ddl(
            "DROP SHARE {share_name:i}",
            {
                "share_name": row["share"],
            },
            condition=self.engine.settings.execute_outbound_share,
        )

        return ResolveResult.DROP

    def set_accounts(self, bp: OutboundShareBlueprint):
        query = self.engine.query_builder()

        query.append(
            "ALTER SHARE {full_name:i} SET ACCOUNTS = {accounts:i}",
            {
                "full_name": bp.full_name,
                "accounts": bp.accounts,
            },
        )

        if bp.share_restrictions is not None:
            query.append_nl("SHARE_RESTRICTIONS = {share_restrictions:b}", {"share_restrictions": bp.share_restrictions})

        self.engine.execute_unsafe_ddl(query, condition=self.engine.settings.execute_outbound_share)

    def create_grant(self, share_name, grant: Grant):
        self.engine.execute_unsafe_ddl(
            "GRANT {privilege:r} ON {on:r} {name:i} TO SHARE {share_name:i}",
            {
                "privilege": grant.privilege,
                "on": grant.on.singular,
                "name": grant.name,
                "share_name": share_name,
            },
            condition=self.engine.settings.execute_outbound_share,
        )

    def drop_grant(self, share_name, grant: Grant):
        self.engine.execute_unsafe_ddl(
            "REVOKE {privilege:r} ON {on:r} {name:i} FROM SHARE {share_name:i}",
            {
                "privilege": grant.privilege,
                "on": grant.on.singular,
                "name": grant.name,
                "share_name": share_name,
            },
            condition=self.engine.settings.execute_outbound_share,
        )

    def get_existing_share_grants(self, share_name):
        grants = []

        cur = self.engine.execute_meta(
            "SHOW GRANTS TO SHARE {share_name:i}",
            {
                "share_name": share_name,
            },
        )

        for r in cur:
            grants.append(
                Grant(
                    privilege=r["privilege"],
                    on=ObjectType[r["granted_on"]],
                    name=build_grant_name_ident(ObjectType[r["granted_on"]], r["name"]),
                )
            )

        return grants



from snowddl.blueprint import NetworkPolicyBlueprint
from snowddl.resolver.abc_resolver import AbstractResolver, ResolveResult, ObjectType


class NetworkPolicyResolver(AbstractResolver):
    skip_on_empty_blueprints = True

    def get_object_type(self) -> ObjectType:
        return ObjectType.NETWORK_POLICY

    def get_existing_objects(self):
        existing_objects = {}

        cur = self.engine.execute_meta("SHOW NETWORK POLICIES")

        for r in cur:
            # SHOW NETWORK POLICIES does not support LIKE filter, so it has to be applied manually in the code
            if self.config.env_prefix and not str(r["name"]).startswith(self.config.env_prefix):
                continue

            existing_objects[r["name"]] = {
                "name": r["name"],
                "entries_in_allowed_ip_list": r["entries_in_allowed_ip_list"],
                "entries_in_blocked_ip_list": r["entries_in_blocked_ip_list"],
                "comment": r["comment"] if r["comment"] else None,
            }

        for name, owner in self.engine.executor.map(self.get_owner_from_grant, existing_objects.keys()):
            if owner != self.engine.context.current_role:
                del existing_objects[name]

        return existing_objects

    def get_owner_from_grant(self, name):
        cur = self.engine.execute_meta(
            "SHOW GRANTS ON NETWORK POLICY {name:i}",
            {
                "name": name,
            },
        )

        for r in cur:
            # Assumption: OWNERSHIP grant always exist, exactly 1 row
            if r["privilege"] == "OWNERSHIP":
                return name, r["grantee_name"]

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(NetworkPolicyBlueprint)

    def create_object(self, bp: NetworkPolicyBlueprint):
        query = self.engine.query_builder()

        query.append(
            "CREATE NETWORK POLICY {name:i}",
            {
                "name": bp.full_name,
            },
        )

        query.append_nl(
            "ALLOWED_IP_LIST = ({allowed_ip_list})",
            {
                "allowed_ip_list": bp.allowed_ip_list,
            },
        )

        if bp.blocked_ip_list:
            query.append_nl(
                "BLOCKED_IP_LIST = ({blocked_ip_list})",
                {
                    "blocked_ip_list": bp.blocked_ip_list,
                },
            )

        if bp.comment:
            query.append_nl(
                "COMMENT = {comment}",
                {
                    "comment": bp.comment,
                },
            )

        self.engine.execute_unsafe_ddl(query, condition=self.engine.settings.execute_network_policy)

        return ResolveResult.CREATE

    def compare_object(self, bp: NetworkPolicyBlueprint, row: dict):
        result = ResolveResult.NOCHANGE

        cur = self.engine.execute_meta(
            "DESC NETWORK POLICY {name:i}",
            {
                "name": bp.full_name,
            },
        )

        existing_allowed_ip_list = []
        existing_blocked_ip_list = []

        for r in cur:
            if r["name"] == "ALLOWED_IP_LIST":
                existing_allowed_ip_list = str(r["value"]).split(",")
            elif r["name"] == "BLOCKED_IP_LIST":
                existing_blocked_ip_list = str(r["value"]).split(",")

        if sorted(bp.allowed_ip_list) != sorted(existing_allowed_ip_list):
            self.engine.execute_unsafe_ddl(
                "ALTER NETWORK POLICY {name:i} SET ALLOWED_IP_LIST = ({allowed_ip_list})",
                {
                    "name": bp.full_name,
                    "allowed_ip_list": bp.allowed_ip_list,
                },
                condition=self.engine.settings.execute_network_policy,
            )

            result = ResolveResult.ALTER

        if sorted(bp.blocked_ip_list) != sorted(existing_blocked_ip_list):
            if bp.blocked_ip_list:
                self.engine.execute_unsafe_ddl(
                    "ALTER NETWORK POLICY {name:i} SET BLOCKED_IP_LIST = ({blocked_ip_list})",
                    {
                        "name": bp.full_name,
                        "blocked_ip_list": bp.blocked_ip_list,
                    },
                    condition=self.engine.settings.execute_network_policy,
                )
            else:
                self.engine.execute_unsafe_ddl(
                    "ALTER NETWORK POLICY {name:i} SET BLOCKED_IP_LIST = ()",
                    {
                        "name": bp.full_name,
                    },
                    condition=self.engine.settings.execute_network_policy,
                )

            result = ResolveResult.ALTER

        if bp.comment != row["comment"]:
            self.engine.execute_unsafe_ddl(
                "ALTER NETWORK POLICY {name:i} SET COMMENT = {comment}",
                {
                    "name": bp.full_name,
                    "comment": bp.comment if bp.comment else None,
                },
                condition=self.engine.settings.execute_network_policy,
            )

            result = ResolveResult.ALTER

        return result

    def drop_object(self, row: dict):
        self.engine.execute_unsafe_ddl(
            "DROP NETWORK POLICY {name:i}",
            {
                "name": row["name"],
            },
            condition=self.engine.settings.execute_network_policy,
        )

        return ResolveResult.DROP



from typing import Union

from .ident import AbstractIdent, DatabaseIdent, SchemaIdent
from .object_type import ObjectType
from ..model import BaseModelWithConfig


class Grant(BaseModelWithConfig):
    privilege: str
    on: ObjectType
    name: AbstractIdent


class AccountGrant(BaseModelWithConfig):
    privilege: str


class FutureGrant(BaseModelWithConfig):
    privilege: str
    on_future: ObjectType
    in_parent: ObjectType
    name: Union[DatabaseIdent, SchemaIdent]



from collections import defaultdict
from fnmatch import translate
from pathlib import Path
from re import compile
from typing import Dict, List, Type, Optional, Union

from snowddl.blueprint import (
    AbstractBlueprint,
    AbstractIdentWithPrefix,
    ObjectType,
    PermissionModel,
    PermissionModelRuleset,
    PermissionModelCreateGrant,
    PermissionModelFutureGrant,
    T_Blueprint,
)


class SnowDDLConfig:
    BUSINESS_ROLE_SUFFIX = "B_ROLE"
    DATABASE_ROLE_SUFFIX = "D_ROLE"
    SCHEMA_ROLE_SUFFIX = "S_ROLE"
    TECHNICAL_ROLE_SUFFIX = "T_ROLE"
    USER_ROLE_SUFFIX = "U_ROLE"
    WAREHOUSE_ROLE_SUFFIX = "W_ROLE"

    OWNER_ROLE_TYPE = "OWNER"
    WRITE_ROLE_TYPE = "WRITE"
    READ_ROLE_TYPE = "READ"
    USAGE_ROLE_TYPE = "USAGE"
    MONITOR_ROLE_TYPE = "MONITOR"

    DEFAULT_PERMISSION_MODEL = "DEFAULT"

    def __init__(self, env_prefix=None):
        self.env_prefix = self._init_env_prefix(env_prefix)

        self.blueprints: Dict[Type[T_Blueprint], Dict[str, T_Blueprint]] = defaultdict(dict)
        self.errors: List[dict] = []

        self.placeholders: Dict[str, Union[bool, float, int, str]] = {}
        self.permission_models: Dict[str, PermissionModel] = self._init_permission_models()

    def get_blueprints_by_type(self, cls: Type[T_Blueprint]) -> Dict[str, T_Blueprint]:
        return self.blueprints.get(cls, {})

    def get_blueprints_by_type_and_pattern(self, cls: Type[T_Blueprint], pattern: str) -> Dict[str, T_Blueprint]:
        pattern = pattern.upper()
        all_blueprints = self.blueprints.get(cls, {})

        # Simple search by blueprint name for patterns without any special chars
        if not any(special_char in pattern for special_char in ["|", "!", "*", "?", "[", "]"]):
            # Add env prefix to pattern IF blueprint type supports it
            if "full_name" in cls.model_fields and issubclass(cls.model_fields["full_name"].annotation, AbstractIdentWithPrefix):
                pattern = f"{self.env_prefix}{pattern}"

            return {pattern: all_blueprints[pattern]} if pattern in all_blueprints else {}

        include_full_names = set()
        exclude_full_names = set()

        for sub_pattern in pattern.split("|"):
            is_exclude = False

            # Exclude sub-pattern
            if sub_pattern.startswith("!"):
                is_exclude = True
                sub_pattern = sub_pattern[1:]

            # Add env prefix to sub-pattern IF blueprint type supports it
            if "full_name" in cls.model_fields and issubclass(cls.model_fields["full_name"].annotation, AbstractIdentWithPrefix):
                sub_pattern = f"{self.env_prefix}{sub_pattern}"

            regexp = compile(translate(sub_pattern))

            if is_exclude:
                exclude_full_names.update(full_name for full_name in all_blueprints if regexp.match(full_name))
            else:
                include_full_names.update(full_name for full_name in all_blueprints if regexp.match(full_name))

        return {
            full_name: bp
            for full_name, bp in all_blueprints.items()
            if full_name in include_full_names and full_name not in exclude_full_names
        }

    def get_placeholder(self, name: str) -> Union[bool, float, int, str]:
        if name not in self.placeholders:
            raise ValueError(f"Unknown placeholder [{name}]")

        return self.placeholders[name]

    def get_permission_model(self, name: Optional[str]) -> PermissionModel:
        if name is None:
            name = self.DEFAULT_PERMISSION_MODEL
        elif name not in self.permission_models:
            raise ValueError(f"Unknown permission model [{name}]")

        return self.permission_models[name]

    def add_blueprint(self, bp: AbstractBlueprint):
        self.blueprints[bp.__class__][str(bp.full_name)] = bp

    def remove_blueprint(self, bp: AbstractBlueprint):
        if str(bp.full_name) not in self.blueprints.get(bp.__class__, {}):
            raise ValueError(f"Blueprint with type [{bp.__class__.__name__}] and name [{bp.full_name}] does not exist in config")

        del self.blueprints[bp.__class__][str(bp.full_name)]

    def add_error(self, path: Path, e: Exception):
        self.errors.append(
            {
                "path": path,
                "error": e,
            }
        )

    def add_placeholder(self, name: str, value: Union[bool, float, int, str]):
        self.placeholders[name] = value

    def add_permission_model(self, name: str, permission_model: PermissionModel):
        self.permission_models[name] = permission_model

    def _init_env_prefix(self, env_prefix):
        if env_prefix:
            env_prefix = str(env_prefix).upper()

            if "__" in env_prefix:
                raise ValueError(f"Env prefix [{env_prefix}] cannot contain [__] double underscore")

            if env_prefix.endswith("_"):
                raise ValueError(f"Env prefix [{env_prefix}] cannot end with [_] underscore")

            return f"{env_prefix}__"

        return ""

    def _init_permission_models(self):
        return {
            self.DEFAULT_PERMISSION_MODEL: PermissionModel(
                ruleset=PermissionModelRuleset.SCHEMA_OWNER,
                owner_create_grants=[
                    PermissionModelCreateGrant(on=ObjectType.FILE_FORMAT),
                    PermissionModelCreateGrant(on=ObjectType.FUNCTION),
                    PermissionModelCreateGrant(on=ObjectType.PROCEDURE),
                    PermissionModelCreateGrant(on=ObjectType.TABLE),
                    PermissionModelCreateGrant(on=ObjectType.VIEW),
                ],
                owner_future_grants=[
                    PermissionModelFutureGrant(privilege="OWNERSHIP", on=ObjectType.ALERT),
                    PermissionModelFutureGrant(privilege="OWNERSHIP", on=ObjectType.DYNAMIC_TABLE),
                    PermissionModelFutureGrant(privilege="OWNERSHIP", on=ObjectType.EVENT_TABLE),
                    PermissionModelFutureGrant(privilege="OWNERSHIP", on=ObjectType.EXTERNAL_TABLE),
                    PermissionModelFutureGrant(privilege="OWNERSHIP", on=ObjectType.FILE_FORMAT),
                    PermissionModelFutureGrant(privilege="OWNERSHIP", on=ObjectType.FUNCTION),
                    PermissionModelFutureGrant(privilege="OWNERSHIP", on=ObjectType.MATERIALIZED_VIEW),
                    PermissionModelFutureGrant(privilege="OWNERSHIP", on=ObjectType.PIPE),
                    PermissionModelFutureGrant(privilege="OWNERSHIP", on=ObjectType.PROCEDURE),
                    PermissionModelFutureGrant(privilege="OWNERSHIP", on=ObjectType.SEQUENCE),
                    PermissionModelFutureGrant(privilege="OWNERSHIP", on=ObjectType.STAGE),
                    PermissionModelFutureGrant(privilege="OWNERSHIP", on=ObjectType.STREAM),
                    PermissionModelFutureGrant(privilege="OWNERSHIP", on=ObjectType.TABLE),
                    PermissionModelFutureGrant(privilege="OWNERSHIP", on=ObjectType.TASK),
                    PermissionModelFutureGrant(privilege="OWNERSHIP", on=ObjectType.VIEW),
                ],
                write_future_grants=[
                    PermissionModelFutureGrant(privilege="READ", on=ObjectType.STAGE),
                    PermissionModelFutureGrant(privilege="WRITE", on=ObjectType.STAGE),
                    PermissionModelFutureGrant(privilege="USAGE", on=ObjectType.STAGE),
                    PermissionModelFutureGrant(privilege="USAGE", on=ObjectType.SEQUENCE),
                    PermissionModelFutureGrant(privilege="INSERT", on=ObjectType.TABLE),
                    PermissionModelFutureGrant(privilege="UPDATE", on=ObjectType.TABLE),
                    PermissionModelFutureGrant(privilege="DELETE", on=ObjectType.TABLE),
                    PermissionModelFutureGrant(privilege="TRUNCATE", on=ObjectType.TABLE),
                ],
                read_future_grants=[
                    PermissionModelFutureGrant(privilege="SELECT", on=ObjectType.DYNAMIC_TABLE),
                    PermissionModelFutureGrant(privilege="SELECT", on=ObjectType.EXTERNAL_TABLE),
                    PermissionModelFutureGrant(privilege="REFERENCES", on=ObjectType.EXTERNAL_TABLE),
                    PermissionModelFutureGrant(privilege="USAGE", on=ObjectType.FILE_FORMAT),
                    PermissionModelFutureGrant(privilege="USAGE", on=ObjectType.FUNCTION),
                    PermissionModelFutureGrant(privilege="SELECT", on=ObjectType.MATERIALIZED_VIEW),
                    PermissionModelFutureGrant(privilege="REFERENCES", on=ObjectType.MATERIALIZED_VIEW),
                    PermissionModelFutureGrant(privilege="USAGE", on=ObjectType.PROCEDURE),
                    PermissionModelFutureGrant(privilege="READ", on=ObjectType.STAGE),
                    PermissionModelFutureGrant(privilege="USAGE", on=ObjectType.STAGE),
                    PermissionModelFutureGrant(privilege="SELECT", on=ObjectType.STREAM),
                    PermissionModelFutureGrant(privilege="SELECT", on=ObjectType.TABLE),
                    PermissionModelFutureGrant(privilege="REFERENCES", on=ObjectType.TABLE),
                    PermissionModelFutureGrant(privilege="SELECT", on=ObjectType.VIEW),
                    PermissionModelFutureGrant(privilege="REFERENCES", on=ObjectType.VIEW),
                ],
            )
        }



from snowddl.blueprint import ObjectType
from snowddl.converter.abc_converter import ConvertResult
from snowddl.converter.abc_schema_object_converter import AbstractSchemaObjectConverter
from snowddl.parser.sequence import sequence_json_schema


class SequenceConverter(AbstractSchemaObjectConverter):
    def get_object_type(self) -> ObjectType:
        return ObjectType.SEQUENCE

    def get_existing_objects_in_schema(self, schema: dict):
        existing_objects = {}

        cur = self.engine.execute_meta(
            "SHOW SEQUENCES IN SCHEMA {database:i}.{schema:i}",
            {
                "database": schema["database"],
                "schema": schema["schema"],
            },
        )

        for r in cur:
            existing_objects[f"{r['database_name']}.{r['schema_name']}.{r['name']}"] = {
                "database": r["database_name"],
                "schema": r["schema_name"],
                "name": r["name"],
                "owner": r["owner"],
                "next_value": r["next_value"],
                "interval": r["interval"],
                "comment": r["comment"] if r["comment"] else None,
            }

        return existing_objects

    def dump_object(self, row):
        data = {
            "interval": row["interval"],
            "comment": row["comment"],
        }

        object_path = (
            self.base_path / self._normalise_name_with_prefix(row["database"]) / self._normalise_name(row["schema"]) / "sequence"
        )

        if data:
            self._dump_file(object_path / f"{self._normalise_name(row['name'])}.yaml", data, sequence_json_schema)
            return ConvertResult.DUMP

        return ConvertResult.EMPTY



from snowddl.blueprint import StreamBlueprint
from snowddl.resolver.abc_schema_object_resolver import AbstractSchemaObjectResolver, ResolveResult, ObjectType
from snowddl.resolver._utils import coalesce


class StreamResolver(AbstractSchemaObjectResolver):
    def get_object_type(self) -> ObjectType:
        return ObjectType.STREAM

    def get_existing_objects_in_schema(self, schema: dict):
        existing_objects = {}

        cur = self.engine.execute_meta(
            "SHOW STREAMS IN SCHEMA {database:i}.{schema:i}",
            {
                "database": schema["database"],
                "schema": schema["schema"],
            },
        )

        for r in cur:
            existing_objects[f"{r['database_name']}.{r['schema_name']}.{r['name']}"] = {
                "database": r["database_name"],
                "schema": r["schema_name"],
                "name": r["name"],
                "object_name": r["table_name"],
                "type": r["type"],
                "mode": r["mode"],
                "comment": r["comment"] if r["comment"] else None,
            }

        return existing_objects

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(StreamBlueprint)

    def create_object(self, bp: StreamBlueprint):
        query = self._build_create_stream_sql(bp)
        self.engine.execute_safe_ddl(query)

        return ResolveResult.CREATE

    def compare_object(self, bp: StreamBlueprint, row: dict):
        is_replace_required = False

        if coalesce(bp.append_only, False) != ("APPEND_ONLY" in row["mode"]):
            is_replace_required = True

        if coalesce(bp.insert_only, False) != ("INSERT_ONLY" in row["mode"]):
            is_replace_required = True

        if is_replace_required:
            query = self._build_create_stream_sql(bp, True)
            self.engine.execute_unsafe_ddl(query)

            return ResolveResult.REPLACE

        if bp.comment != row["comment"]:
            self.engine.execute_safe_ddl(
                "ALTER STREAM {full_name:i} SET COMMENT = {comment}",
                {
                    "full_name": bp.full_name,
                    "comment": bp.comment,
                },
            )

            return ResolveResult.ALTER

        return ResolveResult.NOCHANGE

    def drop_object(self, row: dict):
        self.engine.execute_unsafe_ddl(
            "DROP STREAM {database:i}.{schema:i}.{name:i}",
            {
                "database": row["database"],
                "schema": row["schema"],
                "name": row["name"],
            },
        )

        return ResolveResult.DROP

    def _build_create_stream_sql(self, bp: StreamBlueprint, is_replace=False):
        query = self.engine.query_builder()
        query.append("CREATE")

        if is_replace:
            query.append("OR REPLACE")

        query.append(
            "STREAM {full_name:i}",
            {
                "full_name": bp.full_name,
            },
        )

        # TODO: uncomment when COPY GRANTS is supported for streams
        # https://docs.snowflake.com/en/sql-reference/sql/create-stream.html
        # query.append_nl("COPY GRANTS")

        query.append_nl(
            "ON {object_type:r} {object_name:i}",
            {
                "object_type": bp.object_type.singular,
                "object_name": bp.object_name,
            },
        )

        if bp.append_only is not None:
            query.append_nl(
                "APPEND_ONLY = {append_only:b}",
                {
                    "append_only": bp.append_only,
                },
            )

        if bp.insert_only is not None:
            query.append_nl(
                "INSERT_ONLY = {insert_only:b}",
                {
                    "insert_only": bp.insert_only,
                },
            )

        if bp.show_initial_rows is not None:
            query.append_nl(
                "SHOW_INITIAL_ROWS = {show_initial_rows:b}",
                {
                    "show_initial_rows": bp.show_initial_rows,
                },
            )

        if bp.comment:
            query.append_nl(
                "COMMENT = {comment}",
                {
                    "comment": bp.comment,
                },
            )

        return query



from snowddl.blueprint import TaskBlueprint
from snowddl.resolver.abc_schema_object_resolver import AbstractSchemaObjectResolver, ResolveResult, ObjectType


class TaskResolver(AbstractSchemaObjectResolver):
    def get_object_type(self) -> ObjectType:
        return ObjectType.TASK

    def get_existing_objects_in_schema(self, schema: dict):
        existing_objects = {}

        cur = self.engine.execute_meta(
            "SHOW TASKS IN SCHEMA {database:i}.{schema:i}",
            {
                "database": schema["database"],
                "schema": schema["schema"],
            },
        )

        for r in cur:
            existing_objects[f"{r['database_name']}.{r['schema_name']}.{r['name']}"] = {
                "database": r["database_name"],
                "schema": r["schema_name"],
                "name": r["name"],
                "owner": r["owner"],
                "predecessors": r["predecessors"],
                "comment": r["comment"] if r["comment"] else None,
            }

        return existing_objects

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(TaskBlueprint)

    def create_object(self, bp: TaskBlueprint):
        query = self._build_create_task(bp)

        self.engine.execute_safe_ddl(query)

        self.engine.execute_safe_ddl(
            "COMMENT ON TASK {full_name:i} IS {comment}",
            {
                "full_name": bp.full_name,
                "comment": query.add_short_hash(bp.comment),
            },
        )

        return ResolveResult.CREATE

    def compare_object(self, bp: TaskBlueprint, row: dict):
        query = self._build_create_task(bp)

        if not query.compare_short_hash(row["comment"]):
            self.engine.execute_safe_ddl(query)

            self.engine.execute_safe_ddl(
                "COMMENT ON TASK {full_name:i} IS {comment}",
                {
                    "full_name": bp.full_name,
                    "comment": query.add_short_hash(bp.comment),
                },
            )

            return ResolveResult.REPLACE

        # Set predecessor again if it was dropped
        if bp.after and not row["predecessors"]:
            self.engine.execute_safe_ddl(
                "ALTER TASK {full_name:i} ADD AFTER {after:i}",
                {
                    "full_name": bp.full_name,
                    "after": bp.after,
                },
            )

            return ResolveResult.ALTER

        return ResolveResult.NOCHANGE

    def drop_object(self, row: dict):
        self.engine.execute_safe_ddl(
            "DROP TASK {database:i}.{schema:i}.{name:i}",
            {
                "database": row["database"],
                "schema": row["schema"],
                "name": row["name"],
            },
        )

        return ResolveResult.DROP

    def _build_create_task(self, bp: TaskBlueprint):
        query = self.engine.query_builder()

        query.append(
            "CREATE OR REPLACE TASK {full_name:i}",
            {
                "full_name": bp.full_name,
            },
        )

        if bp.warehouse:
            query.append_nl(
                "WAREHOUSE = {warehouse:i}",
                {
                    "warehouse": bp.warehouse,
                },
            )

        if bp.user_task_managed_initial_warehouse_size:
            query.append_nl(
                "USER_TASK_MANAGED_INITIAL_WAREHOUSE_SIZE = {size:r}",
                {
                    "size": bp.user_task_managed_initial_warehouse_size,
                },
            )

        if bp.schedule:
            query.append_nl(
                "SCHEDULE = {schedule}",
                {
                    "schedule": bp.schedule,
                },
            )

        if bp.config:
            query.append_nl(
                "CONFIG = {config}",
                {
                    "config": bp.config,
                },
            )

        if bp.allow_overlapping_execution:
            query.append_nl(
                "ALLOW_OVERLAPPING_EXECUTION = {allow_overlapping_execution:b}",
                {
                    "allow_overlapping_execution": bp.allow_overlapping_execution,
                },
            )

        if bp.user_task_timeout_ms:
            query.append_nl(
                "USER_TASK_TIMEOUT_MS = {timeout:d}",
                {
                    "timeout": bp.user_task_timeout_ms,
                },
            )

        if bp.suspend_task_after_num_failures:
            query.append_nl(
                "SUSPEND_TASK_AFTER_NUM_FAILURES = {suspend_task_after_num_failures:d}",
                {
                    "suspend_task_after_num_failures": bp.suspend_task_after_num_failures,
                },
            )

        if bp.error_integration:
            query.append_nl(
                "ERROR_INTEGRATION = {error_integration:i}",
                {
                    "error_integration": bp.error_integration,
                },
            )

        if bp.task_auto_retry_attempts:
            query.append_nl(
                "TASK_AUTO_RETRY_ATTEMPTS = {task_auto_retry_attempts:d}",
                {
                    "task_auto_retry_attempts": bp.task_auto_retry_attempts,
                },
            )

        if bp.user_task_minimum_trigger_interval_in_seconds:
            query.append_nl(
                "USER_TASK_MINIMUM_TRIGGER_INTERVAL_IN_SECONDS = {user_task_minimum_trigger_interval_in_seconds:d}",
                {
                    "user_task_minimum_trigger_interval_in_seconds": bp.user_task_minimum_trigger_interval_in_seconds,
                },
            )

        if bp.session_params:
            for param_name, param_value in bp.session_params.items():
                query.append_nl(
                    "{param_name:r} = {param_value:dp}",
                    {
                        "param_name": param_name,
                        "param_value": param_value,
                    },
                )

        if bp.comment:
            query.append_nl(
                "COMMENT = {comment}",
                {
                    "comment": bp.comment,
                },
            )

        if bp.after:
            query.append_nl("AFTER {after:i}", {"after": bp.after})

        if bp.finalize:
            query.append_nl("FINALIZE = {finalize:i}", {"finalize": bp.finalize})

        if bp.when:
            query.append_nl("WHEN {when:r}", {"when": bp.when})

        # Not supported by Snowflake, but documented
        # query.append_nl("COPY GRANTS")

        query.append_nl("AS")
        query.append_nl(bp.body)

        return query



from re import compile

from snowddl.blueprint import (
    Ident,
    DataType,
    HybridTableBlueprint,
    SchemaObjectIdent,
    TableColumn,
    PrimaryKeyBlueprint,
    UniqueKeyBlueprint,
    ForeignKeyBlueprint,
)
from snowddl.resolver.abc_schema_object_resolver import AbstractSchemaObjectResolver, ResolveResult, ObjectType

collate_type_syntax_re = compile(r"^(.*) COLLATE \'(.*)\'$")


class HybridTableResolver(AbstractSchemaObjectResolver):
    skip_on_empty_blueprints = True

    def get_object_type(self) -> ObjectType:
        return ObjectType.HYBRID_TABLE

    def get_existing_objects_in_schema(self, schema: dict):
        existing_objects = {}

        cur = self.engine.execute_meta(
            "SHOW HYBRID TABLES IN SCHEMA {database:i}.{schema:i}",
            {
                "database": schema["database"],
                "schema": schema["schema"],
            },
        )

        for r in cur:
            full_name = f"{r['database_name']}.{r['schema_name']}.{r['name']}"
            existing_objects[full_name] = {
                "database": r["database_name"],
                "schema": r["schema_name"],
                "name": r["name"],
                "owner": r["owner"],
                "comment": r["comment"] if r["comment"] else None,
            }

        return existing_objects

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(HybridTableBlueprint)

    def create_object(self, bp: HybridTableBlueprint):
        common_query = self._build_common_hybrid_table(bp)
        create_query = self.engine.query_builder()

        create_query.append(
            "CREATE HYBRID TABLE {full_name:i}",
            {
                "full_name": bp.full_name,
            },
        )

        create_query.append_nl(common_query)

        self.engine.execute_safe_ddl(create_query)
        self.engine.execute_safe_ddl(
            "COMMENT ON TABLE {full_name:i} IS {comment}",
            {
                "full_name": bp.full_name,
                "comment": common_query.add_short_hash(bp.comment),
            },
        )

        return ResolveResult.CREATE

    def compare_object(self, bp: HybridTableBlueprint, row: dict):
        common_query = self._build_common_hybrid_table(bp)

        if not common_query.compare_short_hash(row["comment"]):
            replace_query = self.engine.query_builder()

            replace_query.append(
                "CREATE OR REPLACE HYBRID TABLE {full_name:i}",
                {
                    "full_name": bp.full_name,
                },
            )

            replace_query.append_nl(common_query)
            replace_query.append_nl(self._build_replace_select_hybrid_table(bp, self._get_existing_columns(bp)))

            self.engine.execute_unsafe_ddl(replace_query, condition=self.engine.settings.execute_replace_table)

            self.engine.execute_unsafe_ddl(
                "COMMENT ON TABLE {full_name:i} IS {comment}",
                {
                    "full_name": bp.full_name,
                    "comment": common_query.add_short_hash(bp.comment),
                },
                condition=self.engine.settings.execute_replace_table,
            )

            return ResolveResult.REPLACE

        return ResolveResult.NOCHANGE

    def drop_object(self, row: dict):
        self.engine.execute_unsafe_ddl(
            "DROP TABLE {database:i}.{schema:i}.{table_name:i}",
            {
                "database": row["database"],
                "schema": row["schema"],
                "table_name": row["name"],
            },
        )

        return ResolveResult.DROP

    def _build_common_hybrid_table(self, bp: HybridTableBlueprint):
        query = self.engine.query_builder()
        query.append("(")

        for idx, c in enumerate(bp.columns):
            query.append_nl(
                "    {comma:r}{col_name:i} {col_type:r}",
                {
                    "comma": "  " if idx == 0 else ", ",
                    "col_name": c.name,
                    "col_type": c.type,
                },
            )

            if c.collate:
                query.append(
                    "COLLATE {collate}",
                    {
                        "collate": c.collate,
                    },
                )

            if c.default is not None:
                query.append(
                    "DEFAULT {default:r}",
                    {
                        "default": self._normalize_bp_default(c.default),
                    },
                )

            if c.not_null:
                query.append("NOT NULL")

            if c.expression:
                query.append(
                    "AS ({expression:r})",
                    {
                        "expression": c.expression,
                    },
                )

            if c.comment:
                query.append(
                    "COMMENT {comment}",
                    {
                        "comment": c.comment,
                    },
                )

        # Primary key
        query.append_nl(
            "    , PRIMARY KEY ({columns:i}) COMMENT {comment}",
            {
                "columns": bp.primary_key,
                "comment": self.get_object_type().name,
            },
        )

        # Unique keys
        if bp.unique_keys:
            for unique_key in bp.unique_keys:
                query.append_nl(
                    "    , UNIQUE ({columns:i}) COMMENT {comment}",
                    {
                        "columns": unique_key,
                        "comment": self.get_object_type().name,
                    },
                )

        # Foreign keys
        if bp.foreign_keys:
            for foreign_key in bp.foreign_keys:
                query.append_nl(
                    "    , FOREIGN KEY ({columns:i}) REFERENCES {ref_table_name:i} ({ref_columns:i}) COMMENT {comment}",
                    {
                        "columns": foreign_key.columns,
                        "ref_table_name": foreign_key.ref_table_name,
                        "ref_columns": foreign_key.ref_columns,
                        "comment": self.get_object_type().name,
                    },
                )

        if bp.indexes:
            for idx in bp.indexes:
                index_name = "__".join(str(part) for part in ["INDEX"] + idx.columns)

                # Snowflake bug: Index name must be used without quotes, otherwise it is actually stored with quotes
                query.append_nl(
                    "    , INDEX {index_name:r} ({columns:i})",
                    {
                        "index_name": index_name,
                        "columns": idx.columns,
                    },
                )

                if idx.include:
                    query.append(
                        "INCLUDE ({include:i})",
                        {
                            "include": idx.include,
                        },
                    )

        query.append_nl(")")

        if bp.comment:
            query.append_nl(
                "COMMENT = {comment}",
                {
                    "comment": bp.comment,
                },
            )

        return query

    def _build_replace_select_hybrid_table(self, bp: HybridTableBlueprint, snow_cols: dict):
        query = self.engine.query_builder()

        query.append("AS")
        query.append_nl("SELECT")

        for idx, c in enumerate(bp.columns):
            col_name = str(c.name)

            if col_name in snow_cols:
                # Column already exist

                if c.type == snow_cols[col_name].type:
                    # Column has the same type
                    query.append_nl(
                        "    {comma:r}{col_name:i} AS {col_name:i}",
                        {
                            "comma": "  " if idx == 0 else ", ",
                            "col_name": c.name,
                        },
                    )
                else:
                    # Column has a different type, add type cast
                    query.append_nl(
                        "    {comma:r}{col_name:i}::{col_type:r} AS {col_name:i}",
                        {
                            "comma": "  " if idx == 0 else ", ",
                            "col_name": c.name,
                            "col_type": c.type,
                        },
                    )
            else:
                # Column does not exist, use default value with type cast
                query.append_nl(
                    "    {comma:r}{col_val}::{col_type:r} AS {col_name:i}",
                    {
                        "comma": "  " if idx == 0 else ", ",
                        "col_name": c.name,
                        "col_type": c.type,
                        "col_val": c.default,
                    },
                )

        query.append_nl(
            "FROM {full_name:i}",
            {
                "full_name": bp.full_name,
            },
        )

        return query

    def _get_existing_columns(self, bp: HybridTableBlueprint):
        existing_columns = {}

        cur = self.engine.execute_meta(
            "DESC TABLE {full_name:i}",
            {
                "full_name": bp.full_name,
            },
        )

        for r in cur:
            m = collate_type_syntax_re.match(r["type"])

            if m:
                dtype = m.group(1)
                collate = m.group(2)
            else:
                dtype = r["type"]
                collate = None

            existing_columns[r["name"]] = TableColumn(
                name=Ident(r["name"]),
                type=DataType(dtype),
                not_null=bool(r["null?"] == "N"),
                default=r["default"] if r["default"] else None,
                expression=r["expression"] if r["expression"] else None,
                collate=collate,
                comment=r["comment"] if r["comment"] else None,
            )

        return existing_columns

    def _normalize_bp_default(self, bp_default):
        if isinstance(bp_default, SchemaObjectIdent):
            # This cannot be formatted properly with double-quotes ("), Snowflake strips it from DEFAULT value returned by DESC TABLE
            return f"{bp_default}.NEXTVAL"

        return bp_default



from typing import Optional, Union

from .data_type import DataType
from .ident import Ident, SchemaObjectIdent
from ..model import BaseModelWithConfig


class DynamicTableColumn(BaseModelWithConfig):
    name: Ident
    comment: Optional[str] = None


class ExternalTableColumn(BaseModelWithConfig):
    name: Ident
    type: DataType
    expr: str
    not_null: bool = False
    comment: Optional[str] = None


class TableColumn(BaseModelWithConfig):
    name: Ident
    type: DataType
    not_null: bool = False
    default: Optional[Union[SchemaObjectIdent, str]] = None
    expression: Optional[str] = None
    collate: Optional[str] = None
    comment: Optional[str] = None


class ViewColumn(BaseModelWithConfig):
    name: Ident
    comment: Optional[str] = None


class NameWithType(BaseModelWithConfig):
    name: Ident
    type: DataType


class ArgumentWithType(NameWithType):
    default: Optional[str] = None


class SearchOptimizationItem(BaseModelWithConfig):
    method: str
    target: Ident



from snowddl.blueprint import AccountIdent, OutboundShareBlueprint, OutboundShareIdent, Grant, ObjectType
from snowddl.parser.abc_parser import AbstractParser, ParsedFile


# fmt: off
outbound_share_json_schema = {
    "type": "object",
    "additionalProperties": {
        "properties": {
            "accounts": {
                "type": "array",
                "items": {
                    "type": "string"
                },
            },
            "share_restrictions": {
                "type": "boolean"
            },
            "grants": {
                "type": "object",
                "additionalProperties": {
                    "type": "array",
                    "items": {
                        "type": "string"
                    },
                    "minItems": 1
                }
            },
            "comment": {
                "type": "string"
            }
        },
        "required": ["grants"],
        "additionalProperties": False
    }
}
# fmt: on


class OutboundShareParser(AbstractParser):
    def load_blueprints(self):
        self.parse_single_file(self.base_path / "outbound_share.yaml", outbound_share_json_schema, self.process_outbound_share)

    def process_outbound_share(self, f: ParsedFile):
        for share_name, share in f.params.items():
            bp = OutboundShareBlueprint(
                full_name=OutboundShareIdent(self.env_prefix, share_name),
                accounts=self.get_share_accounts(share),
                share_restrictions=share.get("share_restrictions", False),
                grants=self.get_share_grants(share),
                comment=share.get("comment"),
            )

            self.config.add_blueprint(bp)

    def get_share_accounts(self, share):
        share_accounts = []

        for account in share.get("accounts", []):
            account_parts = account.split(".")

            if len(account_parts) != 2:
                raise ValueError(
                    f"Invalid outbound share account format [{account}], expected identifier with exactly 2 parts: <organization>.<account>"
                )

            share_accounts.append(AccountIdent(*account_parts))

        return share_accounts

    def get_share_grants(self, share):
        grants = []

        for definition, pattern_list in share["grants"].items():
            on, privileges = definition.upper().split(":")

            for p in privileges.split(","):
                for pattern in pattern_list:
                    blueprints = self.config.get_blueprints_by_type_and_pattern(ObjectType[on].blueprint_cls, pattern)

                    if not blueprints:
                        raise ValueError(f"No {ObjectType[on].plural} matched wildcard grant with pattern [{pattern}]")

                    for object_bp in blueprints.values():
                        grants.append(
                            Grant(
                                privilege=p,
                                on=ObjectType[on],
                                name=object_bp.full_name,
                            )
                        )

        return grants



from snowddl.blueprint import ExternalTableBlueprint
from snowddl.resolver.abc_schema_object_resolver import AbstractSchemaObjectResolver, ResolveResult, ObjectType
from snowddl.resolver._utils import coalesce


class ExternalTableResolver(AbstractSchemaObjectResolver):
    def get_object_type(self) -> ObjectType:
        return ObjectType.EXTERNAL_TABLE

    def get_existing_objects_in_schema(self, schema: dict):
        existing_objects = {}

        cur = self.engine.execute_meta(
            "SHOW EXTERNAL TABLES IN SCHEMA {database:i}.{schema:i}",
            {
                "database": schema["database"],
                "schema": schema["schema"],
            },
        )

        for r in cur:
            full_name = f"{r['database_name']}.{r['schema_name']}.{r['name']}"
            existing_objects[full_name] = {
                "database": r["database_name"],
                "schema": r["schema_name"],
                "name": r["name"],
                "owner": r["owner"],
                "invalid": r["invalid"] == "true",
                "invalid_reason": r["invalid_reason"],
                "stage": r["stage"],
                "location": r["location"],
                "file_format_name": r["file_format_name"],
                "file_format_type": r["file_format_type"],
                "notification_channel": r["notification_channel"],
                "comment": r["comment"] if r["comment"] else None,
            }

        return existing_objects

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(ExternalTableBlueprint)

    def create_object(self, bp: ExternalTableBlueprint):
        query = self._build_create_external_table(bp)

        self.engine.execute_safe_ddl(query)

        self.engine.execute_safe_ddl(
            "COMMENT ON TABLE {full_name:i} IS {comment}",
            {
                "full_name": bp.full_name,
                "comment": query.add_short_hash(bp.comment),
            },
        )

        return ResolveResult.CREATE

    def compare_object(self, bp: ExternalTableBlueprint, row: dict):
        query = self._build_create_external_table(bp)

        if not query.compare_short_hash(row["comment"]) or row["invalid"]:
            self.engine.execute_safe_ddl(query)

            self.engine.execute_safe_ddl(
                "COMMENT ON TABLE {full_name:i} IS {comment}",
                {
                    "full_name": bp.full_name,
                    "comment": query.add_short_hash(bp.comment),
                },
            )

            return ResolveResult.REPLACE

        return ResolveResult.NOCHANGE

    def drop_object(self, row: dict):
        self.engine.execute_unsafe_ddl(
            "DROP EXTERNAL TABLE {database:i}.{schema:i}.{table_name:i}",
            {
                "database": row["database"],
                "schema": row["schema"],
                "table_name": row["name"],
            },
        )

        return ResolveResult.DROP

    def _build_create_external_table(self, bp: ExternalTableBlueprint):
        query = self.engine.query_builder()

        query.append(
            "CREATE OR REPLACE EXTERNAL TABLE {full_name:i}",
            {
                "full_name": bp.full_name,
            },
        )

        if bp.columns:
            query.append_nl("(")

            for idx, c in enumerate(bp.columns):
                query.append_nl(
                    "    {comma:r}{col_name:i} {col_type:r} AS ({col_expr:r})",
                    {
                        "comma": "  " if idx == 0 else ", ",
                        "col_name": c.name,
                        "col_type": c.type,
                        "col_expr": c.expr,
                    },
                )

                if c.not_null:
                    query.append("NOT NULL")

                if c.comment:
                    query.append(
                        "COMMENT {comment}",
                        {
                            "comment": c.comment,
                        },
                    )

            query.append_nl(")")

        if bp.partition_by:
            query.append_nl(
                "PARTITION BY ({partition_by:i})",
                {
                    "partition_by": bp.partition_by,
                },
            )

        if bp.partition_type:
            query.append_nl(
                "PARTITION_TYPE = {partition_type}",
                {
                    "partition_type": bp.partition_type,
                },
            )

        query.append_nl(
            "LOCATION = @{stage_name:i}{path:r}",
            {
                "stage_name": bp.location_stage,
                "path": coalesce(bp.location_path, ""),
            },
        )

        if bp.location_pattern:
            query.append_nl(
                "PATTERN = {pattern}",
                {
                    "pattern": bp.location_pattern,
                },
            )

        query.append_nl(
            "FILE_FORMAT = (FORMAT_NAME = {file_format:i})",
            {
                "file_format": bp.file_format,
            },
        )

        if bp.refresh_on_create is not None:
            query.append_nl(
                "REFRESH_ON_CREATE = {refresh_on_create:b}",
                {
                    "refresh_on_create": bp.refresh_on_create,
                },
            )

        if bp.auto_refresh is not None:
            query.append_nl(
                "AUTO_REFRESH = {auto_refresh:b}",
                {
                    "auto_refresh": bp.auto_refresh,
                },
            )

        if bp.aws_sns_topic:
            query.append_nl(
                "AWS_SNS_TOPIC = {aws_sns_topic}",
                {
                    "aws_sns_topic": bp.aws_sns_topic,
                },
            )

        if bp.table_format:
            query.append_nl(
                "TABLE_FORMAT = {table_format}",
                {
                    "table_format": bp.table_format,
                },
            )

        if bp.integration:
            query.append_nl(
                "INTEGRATION = {integration:i}",
                {
                    "integration": bp.integration,
                },
            )

        if bp.comment:
            query.append_nl(
                "COMMENT = {comment}",
                {
                    "comment": bp.comment,
                },
            )

        query.append_nl("COPY GRANTS")

        return query



from snowddl.blueprint import WarehouseBlueprint, Ident, AccountObjectIdent
from snowddl.parser.abc_parser import AbstractParser, ParsedFile


# fmt: off
warehouse_json_schema = {
    "type": "object",
    "additionalProperties": {
        "type": "object",
        "properties": {
            "type": {
                "type": "string"
            },
            "size": {
                "type": "string"
            },
            "min_cluster_count": {
                "type": "integer"
            },
            "max_cluster_count": {
                "type": "integer"
            },
            "scaling_policy": {
                "type": "string"
            },
            "auto_suspend": {
                "type": "integer"
            },
            "resource_monitor": {
                "type": "string"
            },
            "global_resource_monitor": {
                "type": "string"
            },
            "enable_query_acceleration": {
                "type": "boolean"
            },
            "query_acceleration_max_scale_factor": {
                "type": "integer"
            },
            "warehouse_params": {
                "type": "object",
                "additionalProperties": {
                    "type": ["boolean", "number", "string"]
                }
            },
            "comment": {
                "type": "string"
            }
        },
        "required": ["size"],
        "additionalProperties": False
    }
}
# fmt: on


class WarehouseParser(AbstractParser):
    def load_blueprints(self):
        self.parse_single_file(self.base_path / "warehouse.yaml", warehouse_json_schema, self.process_warehouse)

    def process_warehouse(self, f: ParsedFile):
        for warehouse_name, warehouse in f.params.items():
            resource_monitor = None

            if warehouse.get("resource_monitor"):
                resource_monitor = AccountObjectIdent(self.env_prefix, warehouse["resource_monitor"])

            if warehouse.get("global_resource_monitor"):
                resource_monitor = Ident(warehouse["global_resource_monitor"])

            bp = WarehouseBlueprint(
                full_name=AccountObjectIdent(self.env_prefix, warehouse_name),
                type=warehouse.get("type", "STANDARD"),
                size=warehouse["size"],
                auto_suspend=warehouse.get("auto_suspend", 60),
                min_cluster_count=warehouse.get("min_cluster_count"),
                max_cluster_count=warehouse.get("max_cluster_count"),
                scaling_policy=warehouse.get("scaling_policy"),
                resource_monitor=resource_monitor,
                enable_query_acceleration=warehouse.get("enable_query_acceleration", False),
                query_acceleration_max_scale_factor=warehouse.get("query_acceleration_max_scale_factor"),
                warehouse_params=self.normalise_params_dict(warehouse.get("warehouse_params", {})),
                comment=warehouse.get("comment"),
            )

            self.config.add_blueprint(bp)



from .app import *
from .blueprint import *
from .config import SnowDDLConfig
from .engine import SnowDDLEngine
from .formatter import SnowDDLFormatter
from .query_builder import SnowDDLQueryBuilder
from .settings import SnowDDLSettings
from .error import SnowDDLExecuteError, SnowDDLUnsupportedError
from .parser import *
from .resolver import *
from .version import __version__



from snowddl.blueprint import NetworkRuleBlueprint
from snowddl.resolver.abc_schema_object_resolver import AbstractSchemaObjectResolver, ResolveResult, ObjectType


class NetworkRuleResolver(AbstractSchemaObjectResolver):
    skip_on_empty_blueprints = True

    def get_object_type(self) -> ObjectType:
        return ObjectType.NETWORK_RULE

    def get_existing_objects_in_schema(self, schema: dict):
        existing_objects = {}

        cur = self.engine.execute_meta(
            "SHOW NETWORK RULES IN SCHEMA {database:i}.{schema:i}",
            {
                "database": schema["database"],
                "schema": schema["schema"],
            },
        )

        for r in cur:
            existing_objects[f"{r['database_name']}.{r['schema_name']}.{r['name']}"] = {
                "database": r["database_name"],
                "schema": r["schema_name"],
                "name": r["name"],
                "owner": r["owner"],
                "type": r["type"],
                "mode": r["mode"],
                "comment": r["comment"] if r["comment"] else None,
            }

        return existing_objects

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(NetworkRuleBlueprint)

    def create_object(self, bp: NetworkRuleBlueprint):
        common_query = self._build_common_network_rule_sql(bp)
        create_query = self.engine.query_builder()

        create_query.append(
            "CREATE NETWORK RULE {full_name:i}",
            {
                "full_name": bp.full_name,
            },
        )

        create_query.append(common_query)
        self.engine.execute_safe_ddl(create_query)

        return ResolveResult.CREATE

    def compare_object(self, bp: NetworkRuleBlueprint, row: dict):
        cur = self.engine.execute_meta(
            "DESC NETWORK RULE {full_name:i}",
            {
                "full_name": bp.full_name,
            },
        )

        desc = cur.fetchone()
        is_replace_required = False

        if (
            bp.type != row["type"]
            or bp.mode != row["mode"]
            or bp.value_list != desc["value_list"].split(",")
            or bp.comment != row["comment"]
        ):
            is_replace_required = True

        if not is_replace_required:
            return ResolveResult.NOCHANGE

        common_query = self._build_common_network_rule_sql(bp)
        replace_query = self.engine.query_builder()

        replace_query.append(
            "CREATE OR REPLACE NETWORK RULE {full_name:i}",
            {
                "full_name": bp.full_name,
            },
        )

        replace_query.append(common_query)
        self.engine.execute_safe_ddl(replace_query)

        return ResolveResult.REPLACE

    def drop_object(self, row: dict):
        self.engine.execute_safe_ddl(
            "DROP NETWORK RULE {database:i}.{schema:i}.{name:i}",
            {
                "database": row["database"],
                "schema": row["schema"],
                "name": row["name"],
            },
        )

        return ResolveResult.DROP

    def _build_common_network_rule_sql(self, bp: NetworkRuleBlueprint):
        query = self.engine.query_builder()

        query.append_nl(
            "TYPE = {type}",
            {
                "type": bp.type,
            },
        )

        query.append_nl(
            "VALUE_LIST = ({value_list})",
            {
                "value_list": bp.value_list,
            },
        )

        query.append_nl(
            "MODE = {mode}",
            {
                "mode": bp.mode,
            },
        )

        if bp.comment:
            query.append_nl(
                "COMMENT = {comment}",
                {
                    "comment": bp.comment,
                },
            )

        return query



from snowddl.blueprint import ExternalAccessIntegrationBlueprint, SchemaObjectIdent, AccountObjectIdent, Ident
from snowddl.parser.abc_parser import AbstractParser, ParsedFile


# fmt: off
external_access_integration_json_schema = {
    "type": "object",
    "additionalProperties": {
        "type": "object",
        "properties": {
            "allowed_network_rules": {
                "type": "array",
                "items": {
                    "type": "string"
                },
                "minItems": 1
            },
            "allowed_api_authentication_integrations": {
                "type": "array",
                "items": {
                    "type": "string"
                },
                "minItems": 1
            },
            "allowed_authentication_secrets": {
                "type": "array",
                "items": {
                    "type": "string"
                },
                "minItems": 1
            },
            "enabled": {
                "type": "boolean"
            },
            "comment": {
                "type": "string"
            }
        },
        "required": ["allowed_network_rules"],
        "additionalProperties": False
    }
}
# fmt: on


class ExternalAccessIntegrationParser(AbstractParser):
    def load_blueprints(self):
        self.parse_single_file(
            self.base_path / "external_access_integration.yaml",
            external_access_integration_json_schema,
            self.process_external_access_integration,
        )

    def process_external_access_integration(self, file: ParsedFile):
        for integration_name, integration in file.params.items():
            allowed_network_rules = [
                SchemaObjectIdent(self.env_prefix, *network_rule.split("."))
                for network_rule in integration.get("allowed_network_rules")
            ]
            allowed_api_authentication_integrations = None
            allowed_authentication_secrets = None

            if integration.get("allowed_api_authentication_integrations"):
                allowed_api_authentication_integrations = [
                    Ident(api_integration) for api_integration in integration.get("allowed_api_authentication_integrations")
                ]

            if integration.get("allowed_authentication_secrets"):
                allowed_authentication_secrets = [
                    SchemaObjectIdent(self.env_prefix, *secret_name.split("."))
                    for secret_name in integration.get("allowed_authentication_secrets")
                ]

            bp = ExternalAccessIntegrationBlueprint(
                full_name=AccountObjectIdent(self.env_prefix, integration_name),
                allowed_network_rules=allowed_network_rules,
                allowed_api_authentication_integrations=allowed_api_authentication_integrations,
                allowed_authentication_secrets=allowed_authentication_secrets,
                enabled=integration.get("enabled", True),
                comment=integration.get("comment"),
            )

            self.config.add_blueprint(bp)



from snowddl.blueprint import (
    DatabaseBlueprint,
    Grant,
    BusinessRoleBlueprint,
    Ident,
    ObjectType,
    SchemaBlueprint,
    build_role_ident,
)
from snowddl.parser.abc_parser import AbstractParser, ParsedFile


# fmt: off
business_role_json_schema = {
    "type": "object",
    "additionalProperties": {
        "type": "object",
        "properties": {
            "database_owner": {
                "type": "array",
                "items": {
                    "type": "string"
                }
            },
            "database_write": {
                "type": "array",
                "items": {
                    "type": "string"
                }
            },
            "database_read": {
                "type": "array",
                "items": {
                    "type": "string"
                }
            },
            "schema_owner": {
                "type": "array",
                "items": {
                    "type": "string"
                }
            },
            "schema_read": {
                "type": "array",
                "items": {
                    "type": "string"
                }
            },
            "schema_write": {
                "type": "array",
                "items": {
                    "type": "string"
                }
            },
            "warehouse_usage": {
                "type": "array",
                "items": {
                    "type": "string"
                }
            },
            "warehouse_monitor": {
                "type": "array",
                "items": {
                    "type": "string"
                }
            },
            "technical_roles": {
                "type": "array",
                "items": {
                    "type": "string"
                }
            },
            "tech_roles": {
                "type": "array",
                "items": {
                    "type": "string"
                }
            },
            "global_roles": {
                "type": "array",
                "items": {
                    "type": "string"
                }
            },
            "comment": {
                "type": "string"
            }
        },
        "additionalProperties": False,
    }
}
# fmt: on


class BusinessRoleParser(AbstractParser):
    def load_blueprints(self):
        self.parse_single_file(self.base_path / "business_role.yaml", business_role_json_schema, self.process_business_role)

    def process_business_role(self, file: ParsedFile):
        for business_role_name, business_role in file.params.items():
            business_role_ident = build_role_ident(self.env_prefix, business_role_name, self.config.BUSINESS_ROLE_SUFFIX)

            grants = []

            # Database roles
            for full_schema_name in business_role.get("database_owner", []):
                grants.extend(self.build_database_role_grants(full_schema_name, "OWNER"))

            for full_schema_name in business_role.get("database_read", []):
                grants.extend(self.build_database_role_grants(full_schema_name, "READ"))

            for full_schema_name in business_role.get("database_write", []):
                grants.extend(self.build_database_role_grants(full_schema_name, "WRITE"))

            # Schema roles
            for full_schema_name in business_role.get("schema_owner", []):
                grants.extend(self.build_schema_role_grants(full_schema_name, "OWNER"))

            for full_schema_name in business_role.get("schema_read", []):
                grants.extend(self.build_schema_role_grants(full_schema_name, "READ"))

            for full_schema_name in business_role.get("schema_write", []):
                grants.extend(self.build_schema_role_grants(full_schema_name, "WRITE"))

            # Warehouse roles
            for warehouse_name in business_role.get("warehouse_usage", []):
                grants.append(self.build_warehouse_role_grant(warehouse_name, "USAGE"))

            for warehouse_name in business_role.get("warehouse_monitor", []):
                grants.append(self.build_warehouse_role_grant(warehouse_name, "MONITOR"))

            # Technical roles
            for technical_role_name in business_role.get("technical_roles", []) + business_role.get("tech_roles", []):
                grants.append(
                    Grant(
                        privilege="USAGE",
                        on=ObjectType.ROLE,
                        name=build_role_ident(self.env_prefix, technical_role_name, self.config.TECHNICAL_ROLE_SUFFIX),
                    )
                )

            # Global roles
            for global_role_name in business_role.get("global_roles", []):
                grants.append(
                    Grant(
                        privilege="USAGE",
                        on=ObjectType.ROLE,
                        name=Ident(global_role_name),
                    )
                )

            bp = BusinessRoleBlueprint(
                full_name=business_role_ident,
                grants=grants,
                comment=business_role.get("comment"),
            )

            self.config.add_blueprint(bp)

    def build_database_role_grants(self, full_database_name, grant_type):
        grants = []

        for database_bp in self.config.get_blueprints_by_type_and_pattern(DatabaseBlueprint, full_database_name).values():
            database_permission_model = self.config.get_permission_model(database_bp.permission_model)

            if database_permission_model.ruleset.create_database_owner_role:
                # Databases with "database owner" permission model are assigned directly
                grants.append(
                    Grant(
                        privilege="USAGE",
                        on=ObjectType.ROLE,
                        name=build_role_ident(
                            self.env_prefix,
                            database_bp.full_name.database,
                            grant_type,
                            self.config.DATABASE_ROLE_SUFFIX,
                        ),
                    )
                )
            elif database_permission_model.ruleset.create_schema_owner_role:
                # Databases with "schema owner" permission model are automatically expanded into individual schema roles
                grants.extend(self.build_schema_role_grants(f"{database_bp.full_name.database}.*", grant_type))

        if not grants:
            raise ValueError(f"No {ObjectType.DATABASE.plural} matched wildcard grant with pattern [{full_database_name}]")

        return grants

    def build_schema_role_grants(self, full_schema_name, grant_type):
        grants = []

        for schema_bp in self.config.get_blueprints_by_type_and_pattern(SchemaBlueprint, full_schema_name).values():
            schema_permission_model = self.config.get_permission_model(schema_bp.permission_model)

            if not schema_permission_model.ruleset.create_schema_owner_role and grant_type == self.config.OWNER_ROLE_TYPE:
                raise ValueError(
                    f"Cannot use parameter [schema_owner] for schema [{schema_bp.full_name.database}.{schema_bp.full_name.schema}] due to permission model. "
                    f"Ownership can only be granted on database level via [database_owner] parameter"
                )

            grants.append(
                Grant(
                    privilege="USAGE",
                    on=ObjectType.ROLE,
                    name=build_role_ident(
                        self.env_prefix,
                        schema_bp.full_name.database,
                        schema_bp.full_name.schema,
                        grant_type,
                        self.config.SCHEMA_ROLE_SUFFIX,
                    ),
                )
            )

        if not grants:
            raise ValueError(f"No {ObjectType.SCHEMA.plural} matched wildcard grant with pattern [{full_schema_name}]")

        return grants

    def build_warehouse_role_grant(self, warehouse_name, grant_type):
        return Grant(
            privilege="USAGE",
            on=ObjectType.ROLE,
            name=build_role_ident(self.env_prefix, warehouse_name, grant_type, self.config.WAREHOUSE_ROLE_SUFFIX),
        )



from argparse import ArgumentParser, HelpFormatter
from cryptography.hazmat.primitives import serialization
from importlib.util import module_from_spec, spec_from_file_location
from json import loads as json_loads
from json.decoder import JSONDecodeError
from logging import getLogger, Formatter, StreamHandler
from os import environ, getcwd
from pathlib import Path
from snowflake.connector import connect
from traceback import TracebackException

from snowddl.blueprint import Ident, ObjectType
from snowddl.config import SnowDDLConfig
from snowddl.engine import SnowDDLEngine
from snowddl.parser import default_parser_sequence, PermissionModelParser, PlaceholderParser
from snowddl.resolver import default_resolver_sequence
from snowddl.settings import SnowDDLSettings
from snowddl.version import __version__


class BaseApp:
    application_name = "SnowDDL"
    application_version = __version__

    parser_sequence = default_parser_sequence
    resolver_sequence = default_resolver_sequence

    def __init__(self):
        self.arg_parser = self.init_arguments_parser()
        self.args = self.init_arguments()
        self.logger = self.init_logger()

        self.config_path = self.init_config_path()
        self.config = self.init_config()
        self.settings = self.init_settings()

        self.engine = self.init_engine()

    def init_arguments_parser(self):
        formatter = lambda prog: HelpFormatter(prog, max_help_position=36)

        parser = ArgumentParser(
            prog="snowddl", description="Object management automation tool for Snowflake", formatter_class=formatter
        )

        # Config
        parser.add_argument(
            "-c",
            help="Path to config directory OR name of bundled test config (default: current directory)",
            metavar="CONFIG_PATH",
            default=getcwd(),
        )

        # Auth
        parser.add_argument(
            "-a",
            help="Snowflake account identifier (default: SNOWFLAKE_ACCOUNT env variable)",
            metavar="ACCOUNT",
            default=environ.get("SNOWFLAKE_ACCOUNT"),
        )
        parser.add_argument(
            "-u",
            help="Snowflake user name (default: SNOWFLAKE_USER env variable)",
            metavar="USER",
            default=environ.get("SNOWFLAKE_USER"),
        )
        parser.add_argument(
            "-p",
            help="Snowflake user password (default: SNOWFLAKE_PASSWORD env variable)",
            metavar="PASSWORD",
            default=environ.get("SNOWFLAKE_PASSWORD"),
        )
        parser.add_argument(
            "-k",
            help="Path to private key file (default: SNOWFLAKE_PRIVATE_KEY_PATH env variable)",
            metavar="PRIVATE_KEY",
            default=environ.get("SNOWFLAKE_PRIVATE_KEY_PATH"),
        )

        # Role & warehouse
        parser.add_argument(
            "-r",
            help="Snowflake active role (default: SNOWFLAKE_ROLE env variable)",
            metavar="ROLE",
            default=environ.get("SNOWFlAKE_ROLE"),
        )
        parser.add_argument(
            "-w",
            help="Snowflake active warehouse (default: SNOWFLAKE_WAREHOUSE env variable)",
            metavar="WAREHOUSE",
            default=environ.get("SNOWFLAKE_WAREHOUSE"),
        )

        # Options
        parser.add_argument(
            "--authenticator",
            help="Authenticator: 'snowflake' or 'externalbrowser' (to use any IdP and a web browser) (default: SNOWFLAKE_AUTHENTICATOR env variable or 'snowflake')",
            default=environ.get("SNOWFLAKE_AUTHENTICATOR", "snowflake"),
        )
        parser.add_argument(
            "--passphrase",
            help="Passphrase for private key file (default: SNOWFLAKE_PRIVATE_KEY_PASSPHRASE env variable)",
            default=environ.get("SNOWFLAKE_PRIVATE_KEY_PASSPHRASE"),
        )
        parser.add_argument(
            "--env-prefix",
            help="Env prefix added to global object names, used to separate environments (e.g. DEV, PROD)",
            default=environ.get("SNOWFLAKE_ENV_PREFIX"),
        )
        parser.add_argument(
            "--env-admin-role",
            help="Super administration role which should inherit env prefixed SnowDDL role",
            default=environ.get("SNOWFLAKE_ENV_ADMIN_ROLE"),
        )
        parser.add_argument(
            "--max-workers", help="Maximum number of workers to resolve objects in parallel", default=None, type=int
        )
        parser.add_argument(
            "--query-tag",
            help="Add QUERY_TAG to all queries produced by SnowDDL",
            default=environ.get("SNOWFLAKE_QUERY_TAG"),
        )

        # Logging
        parser.add_argument(
            "--log-level", help="Log level (possible values: DEBUG, INFO, WARNING; default: INFO)", default="INFO"
        )
        parser.add_argument("--show-sql", help="Show executed DDL queries", default=False, action="store_true")

        # Placeholders
        parser.add_argument(
            "--placeholder-path", help="Path to config file with environment-specific placeholders", default=None, metavar=""
        )
        parser.add_argument(
            "--placeholder-values", help="Environment-specific placeholder values in JSON format", default=None, metavar=""
        )

        # Object types
        parser.add_argument(
            "--exclude-object-types", help="Comma-separated list of object types NOT to resolve", default=None, metavar=""
        )
        parser.add_argument(
            "--include-object-types",
            help="Comma-separated list of object types TO resolve, all other types are excluded",
            default=None,
            metavar="",
        )

        # Apply even more unsafe changes
        parser.add_argument(
            "--apply-unsafe",
            help="Additionally apply unsafe changes, which may cause loss of data (ALTER, DROP, etc.)",
            default=False,
            action="store_true",
        )
        parser.add_argument(
            "--apply-replace-table",
            help="Additionally apply REPLACE TABLE when ALTER TABLE is not possible",
            default=False,
            action="store_true",
        )
        parser.add_argument(
            "--apply-all-policy", help="Additionally apply changes to all types of POLICIES", default=False, action="store_true"
        )
        parser.add_argument(
            "--apply-aggregation-policy",
            help="Additionally apply changes to AGGREGATION POLICIES",
            default=False,
            action="store_true",
        )
        parser.add_argument(
            "--apply-masking-policy", help="Additionally apply changes to MASKING POLICIES", default=False, action="store_true"
        )
        parser.add_argument(
            "--apply-projection-policy",
            help="Additionally apply changes to PROJECTION POLICIES",
            default=False,
            action="store_true",
        )
        parser.add_argument(
            "--apply-row-access-policy",
            help="Additionally apply changes to ROW ACCESS POLICIES",
            default=False,
            action="store_true",
        )
        parser.add_argument(
            "--apply-account-params", help="Additionally apply changes to ACCOUNT PARAMETERS", default=False, action="store_true"
        )
        parser.add_argument(
            "--apply-network-policy", help="Additionally apply changes to NETWORK POLICIES", default=False, action="store_true"
        )
        parser.add_argument(
            "--apply-resource-monitor", help="Additionally apply changes to RESOURCE MONITORS", default=False, action="store_true"
        )
        parser.add_argument(
            "--apply-outbound-share", help="Additionally apply changes to OUTBOUND SHARES", default=False, action="store_true"
        )

        # Refresh state of specific objects
        parser.add_argument(
            "--refresh-user-passwords", help="Additionally refresh passwords of users", default=False, action="store_true"
        )
        parser.add_argument(
            "--refresh-future-grants",
            help="Additionally refresh missing grants for existing objects derived from future grants",
            default=False,
            action="store_true",
        )
        parser.add_argument(
            "--refresh-stage-encryption",
            help="Additionally refresh stage encryption parameters for existing external stages",
            default=False,
            action="store_true",
        )
        parser.add_argument(
            "--refresh-secrets",
            help="Additionally refresh secrets",
            default=False,
            action="store_true",
        )

        # Cloning
        parser.add_argument(
            "--clone-table",
            help="Clone all tables from source databases (without env_prefix) to destination databases (with env_prefix)",
            default=False,
            action="store_true",
        )

        # Destroy without env prefix
        parser.add_argument(
            "--destroy-without-prefix", help="Allow {destroy} action without --env-prefix", default=False, action="store_true"
        )

        # Subparsers
        subparsers = parser.add_subparsers(dest="action")
        subparsers.required = True

        subparsers.add_parser("plan", help="Resolve objects, apply nothing, display suggested changes")
        subparsers.add_parser("apply", help="Resolve objects, apply safe changes, display suggested unsafe changes")
        subparsers.add_parser(
            "destroy", help="Drop objects with specified --env-prefix, use it to reset dev and test environments"
        )

        return parser

    def init_arguments(self):
        args = vars(self.arg_parser.parse_args())

        if not self.validate_auth_args(args):
            self.arg_parser.print_help()
            exit(1)

        return args

    def validate_auth_args(self, args):
        if args["authenticator"] == "snowflake":
            if not args["a"] or not args["u"] or (not args["p"] and not args["k"] and "SNOWFLAKE_PRIVATE_KEY" not in environ):
                return False
        elif args["authenticator"] == "externalbrowser":
            if not args["a"] or not args["u"]:
                return False
        elif args["authenticator"] is not None:
            return False
        return True

    def init_logger(self):
        logger = getLogger("snowddl")
        logger.setLevel(self.args.get("log_level", "INFO"))

        formatter = Formatter("%(asctime)s - %(levelname)s - %(message)s")
        formatter.default_msec_format = "%s.%03d"

        handler = StreamHandler()
        handler.setFormatter(formatter)

        logger.addHandler(handler)

        return logger

    def init_config_path(self):
        config_path = Path(self.args["c"])

        if not config_path.exists():
            config_path = Path(__file__).parent.parent / "_config" / self.args["c"]

        if not config_path.exists():
            raise ValueError(f"Config path [{self.args['c']}] does not exist")

        if not config_path.is_dir():
            raise ValueError(f"Config path [{self.args['c']}] is not a directory")

        return config_path.resolve()

    def init_config(self):
        config = SnowDDLConfig(self.args.get("env_prefix"))

        # Placeholders
        placeholder_path = self.get_placeholder_path()
        placeholder_values = self.get_placeholder_values()

        parser = PlaceholderParser(config, self.config_path)
        parser.load_placeholders(placeholder_path, placeholder_values, self.args)

        if config.errors:
            self.output_config_errors(config)
            exit(1)

        # Permission models
        parser = PermissionModelParser(config, self.config_path)
        parser.load_permission_models()

        if config.errors:
            self.output_config_errors(config)
            exit(1)

        # Blueprints
        for parser_cls in self.parser_sequence:
            parser = parser_cls(config, self.config_path)
            parser.load_blueprints()

        if config.errors:
            self.output_config_errors(config)
            exit(1)

        # Custom programmatically generated blueprints and config adjustments
        for module_path in sorted(self.config_path.glob("__custom/*.py")):
            try:
                spec = spec_from_file_location(module_path.name, module_path)

                module = module_from_spec(spec)
                spec.loader.exec_module(module)

                module.handler(config)
            except Exception as e:
                config.add_error(module_path, e)

        if config.errors:
            self.output_config_errors(config)
            exit(1)

        return config

    def init_settings(self):
        settings = SnowDDLSettings()

        if self.args.get("action") in ("apply", "destroy"):
            settings.execute_safe_ddl = True

            if self.args.get("apply_unsafe") or self.args.get("action") == "destroy":
                settings.execute_unsafe_ddl = True

            if self.args.get("apply_replace_table"):
                settings.execute_replace_table = True

            if self.args.get("apply_all_policy"):
                settings.execute_aggregation_policy = True
                settings.execute_masking_policy = True
                settings.execute_projection_policy = True
                settings.execute_row_access_policy = True
                settings.execute_network_policy = True

            if self.args.get("apply_aggregation_policy"):
                settings.execute_aggregation_policy = True

            if self.args.get("apply_masking_policy"):
                settings.execute_masking_policy = True

            if self.args.get("apply_projection_policy"):
                settings.execute_projection_policy = True

            if self.args.get("apply_row_access_policy"):
                settings.execute_row_access_policy = True

            if self.args.get("apply_account_params"):
                settings.execute_account_params = True

            if self.args.get("apply_network_policy"):
                settings.execute_network_policy = True

            if self.args.get("apply_resource_monitor"):
                settings.execute_resource_monitor = True

            if self.args.get("apply_outbound_share"):
                settings.execute_outbound_share = True

        if self.args.get("refresh_user_passwords"):
            settings.refresh_user_passwords = True

        if self.args.get("refresh_future_grants"):
            settings.refresh_future_grants = True

        if self.args.get("refresh_stage_encryption"):
            settings.refresh_stage_encryption = True

        if self.args.get("refresh_secrets"):
            settings.refresh_secrets = True

        if self.args.get("clone_table"):
            if self.args.get("action") != "apply":
                raise ValueError("Argument --clone-table requires action [apply]")

            if not self.args.get("env_prefix"):
                raise ValueError("Argument --clone-table requires argument --env-prefix")

            settings.clone_table = True

        if self.args.get("env_admin_role"):
            settings.env_admin_role = Ident(self.args.get("env_admin_role"))

        if self.args.get("exclude_object_types"):
            try:
                settings.exclude_object_types = [
                    ObjectType[t.strip().upper()] for t in str(self.args.get("exclude_object_types")).split(",")
                ]
            except KeyError as e:
                raise ValueError(f"Invalid object type [{str(e)}]")

        if self.args.get("include_object_types"):
            try:
                settings.include_object_types = [
                    ObjectType[t.strip().upper()] for t in str(self.args.get("include_object_types")).split(",")
                ]
            except KeyError as e:
                raise ValueError(f"Invalid object type [{str(e)}]")

        if self.args.get("max_workers"):
            settings.max_workers = int(self.args.get("max_workers"))

        return settings

    def init_engine(self):
        return SnowDDLEngine(self.get_connection(), self.config, self.settings)

    def get_connection(self):
        options = {
            "account": self.args["a"],
            "user": self.args["u"],
            "role": self.args["r"],
            "warehouse": self.args["w"],
            "application": f"{self.application_name} {self.application_version}",
        }

        if self.args.get("authenticator") == "snowflake":
            key_bytes = None

            if self.args.get("k"):
                key_path = Path(self.args.get("k"))

                if not key_path.is_file():
                    raise ValueError(f"Private key file [{key_path}] does not exist or not a file")

                key_bytes = key_path.read_bytes()
            elif "SNOWFLAKE_PRIVATE_KEY" in environ:
                key_bytes = str(environ["SNOWFLAKE_PRIVATE_KEY"]).encode("utf-8")

            if key_bytes:
                key_password = str(self.args.get("passphrase")).encode("utf-8") if self.args.get("passphrase") else None
                pk = serialization.load_pem_private_key(data=key_bytes, password=key_password)

                options["private_key"] = pk.private_bytes(
                    encoding=serialization.Encoding.DER,
                    format=serialization.PrivateFormat.PKCS8,
                    encryption_algorithm=serialization.NoEncryption(),
                )
            else:
                options["password"] = self.args["p"]
        elif self.args.get("authenticator") == "externalbrowser":
            options["authenticator"] = "externalbrowser"
        else:
            raise ValueError("Only 'snowflake' and 'externalbrowser' authenticators are supported")

        if self.args.get("query_tag"):
            options["session_parameters"] = {
                "QUERY_TAG": self.args.get("query_tag"),
            }

        return connect(**options)

    def execute(self):
        error_count = 0

        with self.engine:
            self.output_engine_context()

            if self.args.get("action") == "destroy":
                if not self.args.get("env_prefix") and not self.args.get("destroy_without_prefix"):
                    raise ValueError("Argument --env-prefix is required for [destroy] action")

                for resolver_cls in self.resolver_sequence:
                    resolver = resolver_cls(self.engine)
                    resolver.destroy()

                    error_count += len(resolver.errors)

                self.engine.context.destroy_role_with_prefix()
            else:
                for resolver_cls in self.resolver_sequence:
                    resolver = resolver_cls(self.engine)
                    resolver.resolve()

                    error_count += len(resolver.errors)

            self.engine.connection.close()
            self.output_engine_stats()
            self.output_engine_warnings()

            if self.args.get("show_sql"):
                self.output_executed_ddl()

            self.output_suggested_ddl()

            if error_count > 0:
                exit(8)

    def output_engine_context(self):
        roles = []

        if self.engine.context.is_account_admin:
            roles.append("ACCOUNTADMIN")

        if self.engine.context.is_sys_admin:
            roles.append("SYSADMIN")

        if self.engine.context.is_security_admin:
            roles.append("SECURITYADMIN")

        self.logger.info(
            f"Snowflake version = {self.engine.context.version} ({self.engine.context.edition.name}), SnowDDL version = {__version__}"
        )
        self.logger.info(f"Account = {self.engine.context.current_account}, Region = {self.engine.context.current_region}")
        self.logger.info(f"Session = {self.engine.context.current_session}, User = {self.engine.context.current_user}")
        self.logger.info(f"Role = {self.engine.context.current_role}, Warehouse = {self.engine.context.current_warehouse}")
        self.logger.info(f"Roles in session = {','.join(roles)}")
        self.logger.info("---")

    def get_placeholder_path(self):
        if self.args.get("placeholder_path"):
            placeholder_path = Path(self.args.get("placeholder_path"))

            if not placeholder_path.is_file():
                raise ValueError(f"Placeholder path [{self.args.get('placeholder_path')}] does not exist or not a file")

            return placeholder_path.resolve()

        return None

    def get_placeholder_values(self):
        if self.args.get("placeholder_values"):
            try:
                placeholder_values = json_loads(self.args.get("placeholder_values"))
            except JSONDecodeError:
                raise ValueError(f"Placeholder values [{self.args.get('placeholder_values')}] are not a valid JSON")

            if not isinstance(placeholder_values, dict):
                raise ValueError(f"Placeholder values [{self.args.get('placeholder_values')}] are not JSON encoded dict")

            for k, v in placeholder_values.items():
                if not isinstance(v, (bool, float, int, str)):
                    raise ValueError(
                        f"Invalid type [{type(v)}] of placeholder [{k.upper()}] value, supported types are: bool, float, int, str"
                    )

            return placeholder_values

        return None

    def output_config_errors(self, config):
        for e in config.errors:
            self.logger.warning(f"[{e['path']}]: {''.join(TracebackException.from_exception(e['error']).format())}")

    def output_engine_stats(self):
        self.logger.info(
            f"Executed {len(self.engine.executed_ddl)} DDL queries, Suggested {len(self.engine.suggested_ddl)} DDL queries"
        )

    def output_engine_warnings(self):
        for object_type, object_names in self.engine.intention_cache.invalid_name_warning.items():
            for name in object_names:
                self.logger.warning(
                    f"Detected {object_type.name} with name [{name}] "
                    f"which does not conform to SnowDDL standards, please rename or drop it manually"
                )

    def output_suggested_ddl(self):
        if self.engine.suggested_ddl:
            print("--- Suggested DDL ---\n")

        for sql in self.engine.suggested_ddl:
            print(f"{sql};\n")

    def output_executed_ddl(self):
        if self.engine.executed_ddl:
            print("--- Executed DDL ---\n")

        for sql in self.engine.executed_ddl:
            print(f"{sql};\n")


def entry_point():
    app = BaseApp()
    app.execute()



from snowddl.blueprint import (
    AccountObjectIdent,
    DynamicTableBlueprint,
    SchemaObjectIdent,
    DynamicTableColumn,
    Ident,
    build_schema_object_ident,
)
from snowddl.parser.abc_parser import AbstractParser, ParsedFile


# fmt: off
dynamic_table_json_schema = {
    "type": "object",
    "properties": {
        "columns": {
            "type": "object",
            "additionalProperties": {
                "type": ["string", "null"]
            }
        },
        "text": {
            "type": "string"
        },
        "target_lag": {
            "type": "string"
        },
        "warehouse": {
            "type": "string",
        },
        "refresh_mode": {
            "type": "string",
        },
        "initialize": {
            "type": "string",
        },
        "cluster_by": {
            "type": "array",
            "items": {
                "type": "string"
            }
        },
        "is_transient": {
            "type": "boolean"
        },
        "retention_time": {
            "type": "integer"
        },
        "depends_on": {
            "type": "array",
            "items": {
                "type": "string"
            },
            "minItems": 1
        },
        "comment": {
            "type": "string"
        },
    },
    "required": ["text", "target_lag", "warehouse"],
    "additionalProperties": False
}
# fmt: on


class DynamicTableParser(AbstractParser):
    unit_singular_to_plural = {
        "second": "seconds",
        "minute": "minutes",
        "hour": "hours",
        "day": "days",
    }

    unit_plural_to_singular = {v: k for k, v in unit_singular_to_plural.items()}

    def load_blueprints(self):
        self.parse_schema_object_files("dynamic_table", dynamic_table_json_schema, self.process_dynamic_table)

    def process_dynamic_table(self, f: ParsedFile):
        column_blueprints = []

        for col_name, col_comment in f.params.get("columns", {}).items():
            column_blueprints.append(
                DynamicTableColumn(
                    name=Ident(col_name),
                    comment=col_comment if col_comment else None,
                )
            )

        bp = DynamicTableBlueprint(
            full_name=SchemaObjectIdent(self.env_prefix, f.database, f.schema, f.name),
            text=f.params["text"],
            columns=column_blueprints if column_blueprints else None,
            target_lag=self.normalize_target_lag(f.params["target_lag"]),
            warehouse=AccountObjectIdent(self.env_prefix, f.params["warehouse"]),
            refresh_mode=f.params.get("refresh_mode").upper() if f.params.get("refresh_mode") else None,
            initialize=f.params.get("initialize").upper() if f.params.get("initialize") else None,
            cluster_by=f.params.get("cluster_by"),
            is_transient=f.params.get("is_transient", False),
            retention_time=f.params.get("retention_time"),
            depends_on=set(
                build_schema_object_ident(self.env_prefix, d, f.database, f.schema) for d in f.params.get("depends_on", [])
            ),
            comment=f.params.get("comment"),
        )

        self.config.add_blueprint(bp)

    def normalize_target_lag(self, target_lag: str):
        if target_lag.upper() == "DOWNSTREAM":
            return "DOWNSTREAM"

        num, _, unit = target_lag.partition(" ")

        num = int(num)
        unit = unit.lower()

        if num > 1 and (unit in self.unit_singular_to_plural):
            unit = self.unit_singular_to_plural[unit]
        elif num == 1 and (unit in self.unit_plural_to_singular):
            unit = self.unit_plural_to_singular[unit]

        return f"{num} {unit}"
